# coding=utf8
from __future__ import print_function
from __future__ import division
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range
from builtins import object
import multiprocessing, time, math, traceback, queue
from multiprocessing import Process
import multiprocessing
import sys

def sstr(obj):
	""" converts any object to str, if necessary encodes unicode chars """
	try:
		return str(obj)
	except UnicodeEncodeError:
		return str(obj).encode('utf-8')

def shorten(text, max_length=140):
	min_length = 0.8*max_length
	shortstr = ""

	if len(text) <= max_length:
		return text

	words = text.split(" ")
	for word in words:
		if len(shortstr) < min_length:
			if len(shortstr) + len(word) + len(" ") <= max_length:
				shortstr += word + " "
			else:
				shortstr = text[0:max_length-2] + " "
	return shortstr[:-1] + "..."


class Event(object):
	"""An event that can have multiple handlers """
	def __init__(self):
		self.handlers = []
	
	def add(self, handler):
		"""Add a handler that is called when the event fires
		Args:
			handler: a callable that is called when the event is fired. 
			The arguments passed to the handler are the sender and additionally arguments that were passed to the fire() function"""
		self.handlers.append(handler)
		return self
	
	def remove(self, handler):
		"""Removes a handler, such that it is no no longer called when the event fires
		Args:
			handler: the handler function to remove"""
		self.handlers.remove(handler)
		return self
	
	def fire(self, sender, earg=None):
		"""Fires the event and calls all registered handlers
		Args:
			sender: the object that fired the event
			earg: Optional argument containing additional information that should be passed to all handlers"""
		for handler in self.handlers:
			handler(sender, earg)
	
	__iadd__ = add
	__isub__ = remove
	__call__ = fire

class Timer(object):
	"""Measures time from when the timer is created and until stop() is called"""
	def __init__(self, description):
		"""Creates a new timer and starts it.
		Args:
			description: a description of what is measured
		"""
		super(Timer, self).__init__()
		self.start = time.time()
		self.description = description

	def stop(self):
		"""Stops the timer. Prints the number of seconds elapsed on console"""
		self.end = time.time()
		self.seconds = self.end-self.start
		print(self.description + " took " + str(self.seconds) + "s.")

class ParMap(object):
	"""Parallel executes a map in several processes"""
	def __init__(self, map_func, num_workers=multiprocessing.cpu_count(), context_func = None):
		"""A parallel implementation of map() that uses multiple processes to divide the compuation.
		Args:
			map_func: the function to apply to each element
			context_func: an optional function that returns a context object (that could be used in a with block). The function is called once for each worker and the context object is passed to the map_func.
			num_workers: the number of worker processes to spawn (defaults to the number of CPU cores available)
		"""
		super(ParMap, self).__init__()
		self.map_func = map_func
		self.context_func = context_func
		self.num_workers = num_workers
		self.request_stop = multiprocessing.Event()
		self._chunksize = 1
		self.chunkseconds = 3.0

	@property
	def chunksize(self):
		return self._chunksize

	def stop(self):
		"""Requests all active workers connected to a running parallel mapping to stop.
		Note that stopping workers is done asynchrously and can take while, even though this function returns immediately."""
		self.request_stop.set()

	def job_is_finished(self, job, timeout = 0.0):
		if "stopped" in job:
			return True
		elif job["worker"] != None:
			return job["worker"]["connection"].poll(timeout)
		else:
			return False

	def map(self, iterable):
		"""Applies the map_func of the ParMap object to all elements in the iterable using parallel worker processes. The result is returned as a generator.
		An optimal chunksize that is submitted to the workers is calculated dynamically.
		Results are calculated on demand, meaning that you have to loop through the generator to continue processing.
		Elements in the iterable have to be pickable to be able to submit them to worker processes.
		
		Args:
			iterable: an iterable (list or generator) that map_func is applied to
		"""
		#initialize
		workers = []
		free_workers = queue.Queue()
		self.request_stop = multiprocessing.Event()
		self._chunksize = 1

		#start up workers
		for i in range(self.num_workers):
			parent_conn, child_conn = multiprocessing.Pipe()
			worker = {} 
			worker["connection"] =  parent_conn
			worker["process"] = Process(target = self._work, args=(child_conn, self.context_func))
			workers.append(worker)
			free_workers.put(worker)
			worker["process"].start()

		#process values
		jobs = {}
		jobid = 0
		batch = []
		last_processing_times = [self.chunkseconds] * 10*self.num_workers #init with chunkseconds, such that intial chunksize is 1
		last_processing_time_pos = 0

		for value in iterable:
			#wait as long as all workers are busy
			while free_workers.empty():
				minjobid = min(jobs.keys())
				jobs[minjobid]["worker"]["connection"].poll(0.1)
				for job in jobs.values():
					if (not "stopped" in job) and job["worker"]["connection"].poll():
						job.update(job["worker"]["connection"].recv())
						free_workers.put(job["worker"])
						job["worker"] = None

			#if job limit reached, wait for leftmost job to finish
			if len(jobs) >= 10*self.num_workers: #do not start jobs for more than 10*workers batches ahead to save memory
				while not (self.job_is_finished(jobs[min(jobs.keys())], timeout = 0.1)):
					pass

			#yield results while leftmost batch is ready
			while len(jobs) > 0 and (self.job_is_finished(jobs[min(jobs.keys())])): 
				minjobid = min(jobs.keys())
				if (not "stopped" in jobs[minjobid]) and jobs[minjobid]["worker"]["connection"].poll(): #FIXME: worker might have new job!!!!
					jobs[minjobid].update(jobs[minjobid]["worker"]["connection"].recv())
					free_workers.put(jobs[minjobid]["worker"])
					jobs[minjobid]["worker"] = None

				#update optimal chunksize based on 10*workers last batch processing times
				last_processing_times[last_processing_time_pos] = (jobs[minjobid]["stopped"] - jobs[minjobid]["started"])/self._chunksize
				last_processing_time_pos = (last_processing_time_pos + 1) % (10*self.num_workers) #rotate through list with last processing times

				avg_processing_time = sum(last_processing_times)/len(last_processing_times)
				desired_chunksize = int(math.ceil(self.chunkseconds/avg_processing_time)) #batch should take chunkseconds s to calculate
				self._chunksize = min(desired_chunksize, max(10,2*self._chunksize)) #double chunksize at most every time (but allow to go to 10 directly in the beginning)
				#print(self._chunksize)

				if "error" in jobs[minjobid]:
					self.request_stop.set()
					#shutdown workers
					for worker in workers:
						try:
							worker["connection"].send(None) #send shutdown command
							worker["connection"].close()
							worker["process"].join()
						except Exception as e:
							pass
						
					ex_type, ex_value, tb_str = jobs[minjobid]["error"]
					message = '%s (in subprocess)\n%s' % (str(ex_value), tb_str)
					raise ex_type(message)

				for r in jobs[minjobid]["results"]:
					yield r
				del jobs[minjobid]

			#start new job if batch full
			batch.append(value)

			if len(batch) >= self._chunksize:
				job = {}
				jobs[jobid] = job
				job["started"] = time.time()
				job["worker"] = free_workers.get()
				job["worker"]["connection"].send(batch)

				batch = []
				jobid += 1

		#wait while all workers busy
		while free_workers.empty():
			minjobid = min(jobs.keys())
			jobs[minjobid]["worker"]["connection"].poll(1)
			for job in jobs.values():
				if (not "stopped" in job) and job["worker"]["connection"].poll():
					job.update(job["worker"]["connection"].recv())
					free_workers.put(job["worker"])
					job["worker"] = None

		#submit last batch
		if len(batch) > 0:
			job = {}
			jobs[jobid] = job
			job["started"] = time.time()
			job["worker"] = free_workers.get()
			job["worker"]["connection"].send(batch)
			batch = []

		#wait for all jobs to finish
		while len(jobs) > 0:
			if (self.job_is_finished(jobs[min(jobs.keys())], timeout = 1.0)):
				minjobid = min(jobs.keys())
				if (not "stopped" in jobs[minjobid]) and jobs[minjobid]["worker"]["connection"].poll():
					jobs[minjobid].update(jobs[minjobid]["worker"]["connection"].recv())
					free_workers.put(jobs[minjobid]["worker"])
					jobs[minjobid]["worker"] = None

				if "error" in jobs[minjobid]:
					self.request_stop.set()
					#shutdown workers
					for worker in workers:
						try:
							worker["connection"].send(None) #send shutdown command
							worker["connection"].close()
							worker["process"].join()
						except Exception as e:
							pass
						
					ex_type, ex_value, tb_str = jobs[minjobid]["error"]
					message = '%s (in subprocess)\n%s' % (str(ex_value), tb_str)
					raise ex_type(message)

				for r in jobs[minjobid]["results"]:
					yield r
				del jobs[minjobid]

		#shutdown workers
		for worker in workers:
			try:
				worker["connection"].send(None) #send shutdown command
				worker["connection"].close()
				worker["process"].join()
			except Exception as e:
				pass

	def _work(self, conn, context_func):
		shutdown_requested = False

		if context_func != None:
			try:
				with context_func() as c:
					while not shutdown_requested and not self.request_stop.is_set():
						if conn.poll(1):
							batch = conn.recv()
							if batch != None:
								self._map_batch(conn, batch, c)
							else:
								shutdown_requested = True
			except Exception as e: #worker not initialized, return error for all incoming jobs
				ex_type, ex_value, tb = sys.exc_info()
				error = (ex_type, ex_value, ''.join(traceback.format_tb(tb)))

				while not shutdown_requested and not self.request_stop.is_set():
					if conn.poll(1):
						batch = conn.recv()
						if batch != None:
							jobinfo = {}
							jobinfo["error"] = error
							jobinfo["stopped"] = time.time()
							conn.send(jobinfo)
							batch = conn.recv()
						else:
							shutdown_requested = True
		else:
			while not shutdown_requested and not self.request_stop.is_set():
				if conn.poll(1):
					batch = conn.recv()
					if batch != None:
						self._map_batch(conn, batch)
					else:
						shutdown_requested = True
		conn.close()

	def _map_batch(self, conn, batch, context = None):
		results = []
		for value in batch:
			if self.request_stop.is_set():
				jobinfo = {}
				jobinfo["error"] = (Exception, "stop requested", "")
				jobinfo["stopped"] = time.time()
				conn.send(jobinfo)
				return
			try:
				if context:
					results.append(self.map_func(value, context))
				else:
					results.append(self.map_func(value))
			except Exception as e:
				ex_type, ex_value, tb = sys.exc_info()
				error = ex_type, ex_value, ''.join(traceback.format_tb(tb))
				
				jobinfo = {}
				jobinfo["error"] = error
				jobinfo["stopped"] = time.time()
				conn.send(jobinfo)
				return

		jobinfo = {}
		jobinfo["stopped"] = time.time()
		jobinfo["results"] = results
		conn.send(jobinfo)



# coding=utf-8
from __future__ import print_function
import unittest, re, operator
import time, random


import pyparade

class TestPyParade(unittest.TestCase):
	"""Uses a comination of map and reduceByKey to calculate occurencies of each word in a text.
	"""
	def test_wordcount(self):
		text = pyparade.Dataset(["abc test abc test test xyz", "abc test2 abc test cde xyz"])
		words = text.flat_map(lambda line: [(word, 1) for word in re.split(" ", line)])
		wordcounts = words.reduce_by_key(operator.add)

		result = wordcounts.collect(name="Counting words")

		correctResult = [("abc", 4), ("test", 4), ("xyz", 2), ("test2", 1), ("cde", 1)]
		self.assertEqual(len(result), len(correctResult))

		for r in result:
			self.assertIn(r, correctResult)

	def test_map(self):
		d = pyparade.Dataset(list(range(0,100000)), name="Numbers with a really extremly unnecessarly long dataset name for no reason")
		
		def f(a):
			#print(str(a) + "->" + str(a+1))
			time.sleep(0.0001)
			return a + 1

		def g(a):
			#print(str(a) + "->" + str(a+1))
			time.sleep(0.001)
			return a + 1

		inc = d.map(f, name="add 1", output_name="Numbers+1").map(g, name="add 1", output_name="Numbers+2").collect()
		equal = [(1 if a == b else 0) for a, b in zip(inc, list(range(2,100002)))]
		self.assertEqual(sum(equal), 100000)

	def test_group(self):
		d = pyparade.Dataset(list(range(0,1000000)))

		def f(a):
			for i in range(0,500):
				random.random()

			return ((a + 1) % 10, a+1)

		def g(a):
			k,values = a
			return (k, sum(values)/len(values))

		result = d.map(f).group_by_key().map(g).collect()

		for i in range(0,10):
			self.assertEqual(result[i][0], i)
			self.assertTrue(abs(result[i][1]-500000)<=10)

	def test_fold(self):
		def f(a):
			#print(str(a) + "->" + str(a+1))
			#time.sleep(0.001)
			for i in range(0,1000):
				random.random()

			return ((a + 1) % 100000, a+1)

		def g(kv):
			k,v = kv
			return v

		result = pyparade.Dataset(list(range(0,100000)), name="Numbers") \
					.map(f, name="calculate", output_name="Key/Value pairs") \
				 	.map(g, name="take value", output_name="Values") \
				 	.fold(0,operator.add,name="sum", output_name="Sum").collect()
		self.assertEqual(result[0], sum(range(1,100001)))

	def test_map(self):
		def slow_generator():
			for i in range(1,100):
				time.sleep(0.1)
				yield i
		
		def f(a):
			#print(str(a) + "->" + str(a+1))
			time.sleep(0.0001)
			return a + 1

		d = pyparade.Dataset(slow_generator(), length=100, name="Slowly generated dataset")
		inc = d.map(f, name="add 1", output_name="Numbers+1").collect()
		equal = [i+1 for i in range(1,100)]
		self.assertEqual(sum(equal), sum(inc))

	def test_batch(self):
		d = pyparade.Dataset(list(range(0,1000)), name="Numbers")

		batches = d.batch(10).collect()
		self.assertEqual(len(batches), 100)
		self.assertEqual(batches[0][0], 0)
		self.assertEqual(batches[99][9], 999)
		print("Test")
		
	def test_error(self):
		def throw_error(value):
			if value == 3:
				raise ValueError(value)
			else:
				return value

		d = pyparade.Dataset([1,2,3,4,5,6,7,8,9], name="Number")

		self.assertRaises(ValueError, d.map(throw_error).collect)



from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import zip
from builtins import object
import threading, multiprocessing, queue, time, collections, sys, traceback

from pyparade.util import ParMap
from pyparade.util.btree import BTree

class Source(object):
	def __init__(self, name="Source", output_name=None):
		super(Source, self).__init__()
		self._stop_requested = threading.Event()
		self.running = threading.Event()
		self.finished = threading.Event()
		self.name = name
		self.output_name = output_name
		self.processes = []
		self.exception = None

	def get_parents(self):
		parents = [self]
		if isinstance(self.source, Source):
			parents.extend(self.source.get_parents())
		return parents

	def _check_stop(self):
		if self._stop_requested.is_set():

			self.finished.set()
			self.running.clear()
			return True
		else:
			return False

	def stop(self):
		if self.running.is_set():
			self._stop_requested.set()

	def __str__(self):
		return self.name

class OutputEndMarker():
	def __init__(self):
		pass

class Operation(Source):
	def __init__(self, source, num_workers=multiprocessing.cpu_count(), context = None, **kwargs):
		super(Operation, self).__init__(**kwargs)
		self.source = source
		self.inbuffer = source._get_buffer()
		self._outbuffer = queue.Queue(10)
		self._last_output = time.time()
		self._outbatch = queue.Queue()
		self.processed = 0
		self.time_started = None
		self.time_finished = None
		self.num_workers = num_workers
		self.context = context

	def __call__(self):
		self.running.set()
		self.time_started = time.time()

		def _run():
			try:
				self.run()
			except BaseException as e:
				ex_type, ex_value, tb = sys.exc_info()
				error = (ex_type, ex_value, ''.join(traceback.format_tb(tb)))
				self.exception = error
			finally:
				self._outputs(OutputEndMarker())


		self._thread = threading.Thread(target=_run, name=str(self))
		self._thread.start()

		while self._thread.is_alive() and not self._check_stop():
			try:
				batch = self._outbuffer.get(True, timeout=1)
				for value in batch:
					yield value
			except Exception as e:
				pass

		self._flush_output(finish=True)

		while not self._outbuffer.empty() and not self._check_stop():
			try:
				batch = self._outbuffer.get(True, timeout=1)
				for value in batch:
					yield value
			except Exception as e:
				pass

		if self._check_stop() and hasattr(self, "pool"):
			self.pool.stop()

		self.time_finished = time.time()
		self.finished.set()
		self.running.clear()

	def _output(self, value):
		self._outputs([value])

	def _outputs(self, values):
		self._outbatch.put(values)
		if time.time() - self._last_output > 0.5:
			self._flush_output()

	def _flush_output(self, finish = False):
		while self._outbuffer.full():
			if self._check_stop():
				return
			time.sleep(1)

		outbatch = []

		finished = False
		batch_element = None
		while not finished and not self._outbatch.empty():
			batch_element = self._outbatch.get()
			if not isinstance(batch_element, OutputEndMarker):
				outbatch.extend(batch_element)
			else:
				outbatch.append(batch_element)
				finished = True

		if finish: #wait for end marker
			while not finished and not self._check_stop():
				try:
					batch_element = self._outbatch.get(timeout=1)
					if not isinstance(batch_element, OutputEndMarker):
						outbatch.extend(batch_element)
					else:
						outbatch.append(batch_element)
						finished = True
				except queue.Empty:
					pass

		self._outbuffer.put(outbatch)
		self._last_output = time.time()

	def _check_stop(self):
		if self._stop_requested.is_set():
			self.time_finished = time.time()
		return super(Operation, self)._check_stop()

	def _generate_input(self):
		for value in self.inbuffer.generate():
			while self._outbuffer.full():
				if self._check_stop():
					raise BufferError("stop requested")
				time.sleep(1)
			yield value


class MapOperation(Operation):
	def __init__(self, source, map_func, num_workers = multiprocessing.cpu_count(), context = None, name = "Map", **kwargs):
		super(MapOperation, self).__init__(source, num_workers, context, name = name, **kwargs)
		self.map_func = map_func
		self.pool = None #multiprocessing.Pool(num_workers, maxtasksperchild = 1000, initializer = initializer)

	def run(self):
		self.pool = ParMap(self.map_func, num_workers = self.num_workers, context_func = self.context)
		#map
		result = []
		for response in self.pool.map(self._generate_input()):
			if self._check_stop():
				self.pool.stop()
				return

			self.processed += 1
			self._output(response)

class FlatMapOperation(MapOperation):
	"""Calls the map function for every value in the dataset and then flattens the result"""
	def __init__(self, source, map_func, num_workers=multiprocessing.cpu_count(), context = None, name = "FlatMap", **kwargs):
		super(FlatMapOperation, self).__init__(source, map_func, num_workers, context, name = name, **kwargs)

	def run(self):
		self.pool = ParMap(self.map_func, num_workers = self.num_workers, context_func = self.context)
		#map
		result = []
		for response in self.pool.map(self._generate_input()):
			if self._check_stop():
				self.pool.stop()
				return

			self.processed += 1
			#flatten result
			for r in response: 
				self._output(r)

class BatchOperation(Operation):
	def __init__(self, source, batch_size, name = "Batch", **kwargs):
		"""An operation that returns the elements of the source in batches of n elements.

		Args:
			source: The `pyparade.Dataset`to use as the source of elements
			batch_size: The number of elements in each batch 
			**kwargs: Other arguments are passed on to `pyparade.operations.Operation.__init__`
		"""
		super(BatchOperation, self).__init__(source, name = name, **kwargs)
		if batch_size >= 1:
			self.batch_size = batch_size
		else:
			raise ValueError("Illegal batch size")

	def run(self):
		#pack batches
		n = 0
		batch = []
		for element in self._generate_input():
			if self._check_stop():
				self.pool.stop()
				return

			batch.append(element)
			n = n + 1
			self.processed += 1
			
			if n >= self.batch_size: #batch full
				self._output(batch)
				n = 0
				batch = []
		
		if n > 0:
			self._output(batch)

		
class GroupByKeyOperation(Operation):
	"""Groups the key/value pairs and yields tuples (key, [list of values])"""
	def __init__(self, source, partly = False, num_workers=multiprocessing.cpu_count(), name = "GroupByKey", **kwargs):
		super(GroupByKeyOperation, self).__init__(source, num_workers, name = name, **kwargs)
		self.partly = partly

	def run_partly(self, chunksize):
		tree = BTree(chunksize, None, None)

		items_since_last_group = 0
		last_key = None
		for k, v in self._generate_input():
			if self._check_stop():
				return

			if k in tree:
				tree[k].append(v)
			else:
				tree[k] = [v]
			self.processed += 1
			items_since_last_group += 1

			if items_since_last_group >= (chunksize*self.num_workers) and k != last_key: #only output elements periodically and once key changes
				keyleafs = [l for l in tree.get_leafs()]
				for key, leaf in keyleafs:
					for k, v in zip(leaf.keys, leaf.values):
						self._output((k,v))
				tree = BTree(chunksize, None, None)
				items_since_last_group = 0

			last_key = k

		keyleafs = [l for l in tree.get_leafs()]
		for key, leaf in keyleafs:
			for k, v in zip(leaf.keys, leaf.values):
				self._output((k,v))

	def run(self, chunksize=10):
		if self.partly:
			return self.run_partly(chunksize)

		tree = BTree(chunksize, None, None)

		for k, v in self._generate_input():
			if self._check_stop():
				return

			if k in tree:
				tree[k].append(v)
			else:
				tree[k] = [v]
			self.processed += 1

		keyleafs = [l for l in tree.get_leafs()]
		for key, leaf in keyleafs:
			for k, v in zip(leaf.keys, leaf.values):
				self._output((k,v))

class ReduceByKeyOperation(Operation):
	"""Reduces the dataset by grouping the key/value pairs by key and applying the reduce_func to the values of each group"""
	def __init__(self, source, reduce_func, num_workers=multiprocessing.cpu_count(), name = "ReduceByKey", **kwargs):
		super(ReduceByKeyOperation, self).__init__(source, num_workers, name = name, **kwargs)
		self.reduce_func = reduce_func

	def run(self, chunksize=10):
		tree = BTree(chunksize, None, None)

		for k, v in self._generate_input():
			if self._check_stop():
				return

			if k in tree:
				tree[k] = self.reduce_func(tree[k], v)
			else:
				tree[k] = v
			self.processed += 1

		keyleafs = [l for l in tree.get_leafs()]
		for key, leaf in keyleafs:
			for k, v in zip(leaf.keys, leaf.values):
				self._output((k,v))

class FoldOperation(Operation):
	"""Folds the dataset using a combine function"""
	def __init__(self, source, zero_value, fold_func, num_workers=multiprocessing.cpu_count(), context = None, name = "Fold", **kwargs):
		super(FoldOperation, self).__init__(source, num_workers, context, name = name, **kwargs)
		self.pool = None #ParMap(self._fold_batch, num_workers = num_workers) #futures.ThreadPoolExecutor(num_workers)
		self.zero_value = zero_value
		self.fold_func = fold_func

	def _generate_input_batches(self, chunksize):
		batch = []
		for value in self._generate_input():
			batch.append(value)

			if len(batch) == chunksize:
				yield batch
				batch = []
		yield batch

	def _fold_batch(self, batch):
		result = self.zero_value
		for value in batch:
			result = self.fold_func(result, value)
		return result

	def run(self, chunksize = 10):
		self.pool = ParMap(self._fold_batch, num_workers = self.num_workers, context_func = self.context)
		result = []

		for response in self.pool.map(self._generate_input_batches(chunksize = chunksize)):
			if self._check_stop():
				self.pool.stop()
				return

			result.append(response)

			if len(result) == chunksize:
				result = [self._fold_batch(result)]

			self.processed += 1
		
		self._output(self._fold_batch(result))




from builtins import zip
from builtins import map
from builtins import range
import random
import unittest, time, threading

from pyparade.util import Event, ParMap, Timer

class TestEvent(unittest.TestCase):
	def test_fire_event(self):
		e = Event()
		self.fired = False
		
		def f(self, sender):
			global fired
			self.fired = True

		e += f #add listener
		e(self) #fire Event

		self.assertEqual(True, self.fired)

	def test_remove_listener(self):
		e = Event()
		self.fired = False
		
		def f(self, sender):
			global fired
			self.fired = True

		e += f #add listener
		e -= f #remove listener
		e(self) #fire Event

		self.assertEqual(False, self.fired)

class TestParMap(unittest.TestCase):
	"""Tests parallelized map"""

	def test_stop(self):
		def f(a):
			for i in range(0,1000):
				random.random()

			return ((a + 1) % 100000, a+1)

		def m():
			p.map(list(range(1000000)))

		p = ParMap(f)
		t = threading.Thread(target=m)
		t.start()
		time.sleep(10)
		p.stop()
		time.sleep(10)
		self.assertTrue(not t.is_alive())

	def test_plus_one(self):
		def f(a):
			for i in range(0,1000):
				random.random()

			return ((a + 1) % 100000, a+1)

		p = ParMap(f)
		t_par = Timer("parmap")
		#for r in p.map(range(1000000)):
		#	print(r)
		calculated_values = [v for v in p.map(list(range(1000000)))]
		t_par.stop()
		t_map = Timer("map")
		correct_values = list(map(f,list(range(1000000))))
		t_map.stop()
		self.assertLessEqual(t_par.seconds, 0.8*t_map.seconds)
		for calculated_value, correct_value in zip(calculated_values, correct_values):
			self.assertEqual(correct_value, calculated_value)
		self.assertEqual(len(correct_values), len(calculated_values))

	def test_vary_time(self):
		def f(a):
			for i in range(0,random.randint(1,10000)):
				random.random()

			return ((a + 1) % 100000, a+1)

		p = ParMap(f)
		t_par = Timer("parmap")
		calculated_values = [v for v in p.map(list(range(100000)))]
		t_par.stop()
		t_map = Timer("map")
		correct_values = list(map(f,list(range(100000))))
		t_map.stop()
		for calculated_value, correct_value in zip(calculated_values, correct_values):
			self.assertEqual(correct_value, calculated_value)
		self.assertEqual(len(correct_values), len(calculated_values))




from builtins import zip
from builtins import str
from builtins import range
from builtins import object
import os, json, hashlib
from abc import ABCMeta, abstractmethod

from pyparade.util import sstr, Event
from future.utils import with_metaclass

#--------------------
#B+ Tree implemention
#--------------------

class BTreeNode(with_metaclass(ABCMeta, object)):
	def __init__(self, parent):
		self.parent = parent

	def _get_page_size(self):
		return self.parent._get_page_size()

	def _get_leaffactory(self):
		return self.parent._get_leaffactory()

	def _get_key_class(self):
		return self.parent._get_key_class()

	def _get_value_class(self):
		return self.parent._get_value_class()

	@abstractmethod
	def insert(self, key, value):
		pass

	@abstractmethod
	def remove(self, key):
		pass

	@abstractmethod
	def split(self):
		pass

	@abstractmethod
	def merge(self, leaf):
		pass

	def search(self, key):
		if isinstance(self, BTreeLeafNode):
			return self
		elif len(self.keys) > 0:
			if key < self.keys[0]:
				return self.before(0).search(key)
			elif key >= self.keys[-1]:
				return self.after(len(self.keys)-1).search(key)
			else:
				i = 0
				while not(self.keys[i] <= key < self.keys[i+1]):
					i += 1
				return self.after(i).search(key)
		else:
			return self.childs[0].search(key)

	page_size = property(_get_page_size)
	leaffactory = property(_get_leaffactory)
	key_class = property(_get_key_class)
	value_class = property(_get_value_class)

	@abstractmethod
	def to_JSON(self):
		pass

	@staticmethod
	def from_JSON(parent, nodejson):
		if "childs" in nodejson:
			return BTreeInteriorNode.from_JSON(parent, nodejson)
		else:
			return parent.leaffactory.construct_leaf_from_JSON(parent, nodejson)


class BTreeLeafNode(BTreeNode):
	def __init__(self, parent):
		super(BTreeLeafNode, self).__init__(parent)

	def get_leafs(self, key = None):
		return [(key, self)]

class BTreeLeafFactory(object):
	"""abstract class for btree leaf allocation"""
	def __init__(self):
		super(BTreeLeafFactory, self).__init__()

	@abstractmethod
	def allocate_leaf(self, parent):
		pass

	@abstractmethod
	def deallocate_leaf(self, leaf):
		pass
		
class BTreeMemoryLeafNode(BTreeLeafNode):
	"""a B+Tree leaf that stores the contents in memory"""
	def __init__(self, parent):
		super(BTreeMemoryLeafNode, self).__init__(parent)
		self.keys = []
		self.values = []

	def __getitem__(self, key):
		return self.values[self.keys.index(key)]

	def __setitem__(self, key, value):
		try:
			self.remove(key)
		except Exception as e:
			pass
		self.insert(key, value)

	def __contains__(self, key):
		return key in self.keys

	def __len__(self):
		return len(self.keys)

	def __str__(self):
		return str([str(key) for key in self.keys])

	def to_JSON(self):
		leafjson = {}
		if parent.key_class in [int, dict, list, str, str]:
			leafjson["keys"] = self.keys
		else:
			leafjson["keys"] = [key.to_JSON() for key in self.keys]

		if parent.value_class in [int, dict, list, str, str]:
			leafjson["values"] = self.values
		else:
			leafjson["values"] = [value.to_JSON() for value in self.values]

		return leafjson

	@staticmethod
	def from_JSON(parent, nodejson):
		node = BTreeMemoryLeafNode(parent)
		if parent.key_class in [int, dict, list, str, str]:
			node.keys = nodejson["keys"]
		else:
			node.keys = [parent.key_class.from_JSON(keyjson) for keyjson in nodejson["keys"]]

		if parent.value_class in [int, dict, list, str, str]:
			node.values = nodejson["values"]
		else:
			node.values = [parent.value_class.from_JSON(childjson) for childjson in nodejson["values"]]
		
		return node

	def insert(self, key, value):
		index = 0
		while index < len(self) and self.keys[index] <= key:
			index += 1
		self.keys.insert(index, key)
		self.values.insert(index, value)

		if len(self) > self.page_size:
			key, newnode = self.split()
			self.parent.insert(key, newnode)

	def remove(self, key):
		index = 0
		while index < len(self) and self.keys[index] != key:
			index += 1

		if index == len(self):
			raise KeyError("Key not found")

		if index == 0 and len(self) > 1:
			oldkey = self.keys[0]
			newkey = self.keys[1]
			parent = self.parent
			while not(isinstance(parent, BTree)):
				if oldkey in parent.keys:
					keyindex = parent.keys.index(oldkey)
					del parent.keys[keyindex]
					parent.keys.insert(keyindex, newkey)
				parent = parent.parent

		del self.keys[index]
		del self.values[index]	

		#balance
		if len(self) < self.page_size//2:
			index = self.parent.childs.index(self)
			node = None
			#borrow key from neighbor, try right first
			index = self.parent.childs.index(self)
			if index+1 < len(self.parent.childs) and len(self.parent.childs[index+1]) > self.page_size // 2:
				node = self.parent.childs[index+1]
				key = self.parent.keys[index]
				self.borrowRight(key,node)
				return
			if index-1 >= 0 and len(self.parent.childs[index-1]) > self.page_size // 2:
				node = self.parent.childs[index-1]
				key = self.parent.keys[index-1]
				self.borrowLeft(key,node)
				return

			#merge with neighbor, try right first
			if index+1 < len(self.parent.childs):
				node = self.parent.childs[index+1] #try right first
				self.merge(node)
				self.parent.remove(node)
			elif index-1 >= 0:
				node = self.parent.childs[index-1] #merge with left instead
				node.merge(self)
				self.parent.remove(self)

	def borrowLeft(self, key, leaf):
		pairs = list(zip(leaf.keys, leaf.values))
		pairs.reverse() #to pop left first
		oldseperator = key
		newseperator = key
		while not(len(self) >= self.page_size // 2) or len(self) < len(leaf)-1:
			key, value = pairs.pop(0)
			self.keys.insert(0,key)
			self.values.insert(0,value)
			del leaf.values[-1]
			del leaf.keys[-1]
			newseperator = self.keys[0]
		leaf.parent.replaceKey(oldseperator,newseperator)

	def borrowRight(self, key, leaf):
		pairs = list(zip(leaf.keys, leaf.values))
		oldseperator = key
		newseperator = key
		while not(len(self) >= self.page_size // 2) or len(self) < len(leaf)-1:
			key, value = pairs.pop(0)
			self.keys.append(key)
			self.values.append(value)
			del leaf.values[0]
			del leaf.keys[0]
			newseperator = leaf.keys[0]
		leaf.parent.replaceKey(oldseperator,newseperator)

	def split(self):
		center = len(self)//2
		key = self.keys[center]

		newnode = self.leaffactory.allocate_leaf(None)
		newnode.keys = self.keys[center:]
		newnode.values = self.values[center:]
		self.keys = self.keys[:center]
		self.values = self.values[:center]

		return key, newnode

	def merge(self, leaf):
		for (key, value) in zip(leaf.keys, leaf.values):
			self.insert(key, value)

class BTreeFileLeafFactory(BTreeLeafFactory):
	"""Constructs and destructs file B+ tree file leaves"""
	def __init__(self, path, extension):
		super(BTreeFileLeafFactory, self).__init__()
		self.path = path
		self.extension = extension
		self.allocated_leaves = []
		self.allocated_pages = []
		self.page_changed = Event()
		self.page_changed += self.handle_page_changed
		self.tracking_objects = []

	def get_filename(self, page):
		return self.path + str(page) + self.extension

	def get_page_no(self, leaf):
		index = self.allocated_leaves.index(leaf)
		return self.allocated_pages[index]

	def allocate_leaf(self, parent):
		#get a free page number
		page = 0
		while page in self.allocated_pages:
			page += 1
		leaf = BTreeFileLeafNode(parent, self.get_filename(page))
		leaf.isloaded = True
		self.allocated_leaves.append(leaf)
		self.allocated_pages.append(page)

		return leaf

	def deallocate_leaf(self, leaf):
		index = self.allocated_leaves.index(leaf)
		page = self.allocated_pages[index]
		os.remove(self.get_filename(self.allocated_pages[index]))
		del self.allocated_leaves[index]
		del self.allocated_pages[index]
		[tracking.removed_pages.append(page) for tracking in self.tracking_objects]

	def cleanup(self):
		[leaf.unload() for leaf in self.allocated_leaves]

	def track_changes(self):
		tracking_object = BTreeFileLeafNodeChanges(self)
		self.tracking_objects.append(tracking_object)
		return tracking_object

	def handle_page_changed(self, sender, event):
		if sender in self.allocated_leaves:
			page = self.allocated_pages[self.allocated_leaves.index(sender)]
			[tracking.changed_pages.append(page) for tracking in self.tracking_objects]

	def construct_leaf_from_JSON(self, parent, leafjson):
		page = leafjson["page"]
		if page in self.allocated_pages:
			raise ValueError("Page already allocated")
		leaf = BTreeFileLeafNode(parent, self.get_filename(page))

		self.allocated_leaves.append(leaf)
		self.allocated_pages.append(page)

		if "count" in leafjson:
			leaf.count = leafjson["count"]

		if "hash" in leafjson:
			leaf.hash = leafjson["hash"]	

		return leaf

class BTreeFileLeafNodeChanges(object):
	"""documents changes to file leafs of a file leaf factory"""
	def __init__(self, leaffactory):
		super(BTreeFileLeafNodeChanges, self).__init__()
		self.leaffactory = leaffactory
		self.changed_pages = []
		self.removed_pages = []

	def stop_tracking(self):
		self.leaffactory.tracking_objects.remove(self)
		


class BTreeFileLeafNode(BTreeMemoryLeafNode):
	"""a B+Tree leaf that stores the contents in a file on disk"""
	def __init__(self, parent, filename):
		super(BTreeFileLeafNode, self).__init__(parent)
		self.keys = []
		self.values = []
		self.filename = filename
		self.isloaded  = False
		self.count = 0
		self.hash = ""

	def __getitem__(self, key):
		self.ensure_load()
		return self.values[self.keys.index(key)]

	def __setitem__(self, key, value):
		try:
			self.remove(key)
		except Exception as e:
			pass
		self.insert(key, value)

	def __contains__(self, key):
		self.ensure_load()
		return key in self.keys

	def ensure_load(self):
		if not(self.isloaded):
			if os.path.isfile(self.filename):
				with open(self.filename) as f:
					leafjson = json.loads(f.read())
				if leafjson["version"] != 1:
					raise IOException("Version of B+ tree page file " + sstr(self.filename) + " is not compatible.")

				if self.key_class in [int, dict, list, str, str]:
					self.keys = leafjson["keys"]
				else:
					self.keys = [self.key_class.from_JSON(keyjson) for keyjson in leafjson["keys"]]

				if self.value_class in [int, dict, list, str, str]:
					self.values = leafjson["values"]
				else:
					self.values = [self.value_class.from_JSON(value_json) for value_json in leafjson["values"]]
				
				self.isloaded  = True
				self.count = len(self.keys)
			else:
				self.isloaded  = True
				self.count = 0

	def write(self):
		if self.isloaded:
			self.count = len(self.keys)
			leafjson = {}
			leafjson["version"] = 1
			if len(self) > 0:
				if self.key_class in [int, dict, list, str, str]:
					leafjson["keys"] = self.keys
				else:
					leafjson["keys"] = [key.to_JSON() for key in self.keys]

				if self.value_class in [int, dict, list, str, str]:
					leafjson["values"] = self.values
				else:
					leafjson["values"] = [value.to_JSON() for value in self.values]
			else:
				leafjson["keys"] = []
				leafjson["values"] = []

			s = json.dumps(leafjson, indent=3)
			self.hash = sstr(hashlib.md5(s.encode('utf-8')).hexdigest())
			with open(self.filename, 'w') as f:
				f.write(s + "\n")
			self.leaffactory.page_changed(self, None)

	def to_JSON(self):
		return {"page": self.leaffactory.allocated_pages[self.leaffactory.allocated_leaves.index(self)],
				"count": self.count,
				"hash": self.hash}

	def unload(self):
		self.isloaded  = False
		self.keys = []
		self.values = []

	def __len__(self):
		if self.isloaded:
			self.count = len(self.keys)
		return self.count

	def __str__(self):
		self.ensure_load()
		return str([str(key) for key in self.keys])

	def insert(self, key, value):
		self.ensure_load()
		super(BTreeFileLeafNode, self).insert(key, value)
		self.write()

	def remove(self, key):
		self.ensure_load()
		super(BTreeFileLeafNode, self).remove(key)
		self.write()

	def borrowLeft(self, key, leaf):
		self.ensure_load()
		leaf.ensure_load()
		super(BTreeFileLeafNode, self).borrowLeft(key, leaf)
		self.write()
		leaf.write()

	def borrowRight(self, key, leaf):
		self.ensure_load()
		leaf.ensure_load()
		super(BTreeFileLeafNode, self).borrowRight(key, leaf)
		self.write()
		leaf.write()

	def split(self):
		self.ensure_load()
		center = len(self)//2
		key = self.keys[center]

		newnode = self.leaffactory.allocate_leaf(self.parent)
		newnode.keys = self.keys[center:]
		newnode.values = self.values[center:]
		self.keys = self.keys[:center]
		self.values = self.values[:center]

		self.write()
		newnode.write()
		return key, newnode

	def merge(self, leaf):
		leaf.ensure_load()
		super(BTreeFileLeafNode, self).merge(leaf)

class BTreeMemoryLeafFactory(BTreeLeafFactory):
	"""docstring for BTreeMemoryLeafFactory"""
	def __init__(self):
		super(BTreeMemoryLeafFactory, self).__init__()

	def allocate_leaf(self, parent):
		return BTreeMemoryLeafNode(parent)

	def deallocate_leaf(self, leaf):
		pass


class BTree(object):
	def __init__(self, page_size, key_class, value_class, leaffactory = BTreeMemoryLeafFactory(), init = True):
		super(BTree, self).__init__()
		self._page_size = page_size
		self._leaffactory = leaffactory
		self._key_class = key_class
		self._value_class = value_class
		self.root = None

		if init:
			self.root = BTreeInteriorNode(None)
			self.root.isroot = True
			self.root.childs.append(leaffactory.allocate_leaf(self.root))
			self.root.parent = self

	def to_JSON(self):
		return {"version": 1,\
				"page_size": self.page_size,\
				"root": self.root.to_JSON()}

	@staticmethod
	def from_JSON(btreejson, key_class, value_class, leaffactory =  BTreeMemoryLeafFactory()):
		if btreejson["version"] != 1:
			raise IOException("Version of B+ tree JSON is not compatible.")
		btree = BTree(btreejson["page_size"], key_class, value_class, leaffactory, init = False)
		btree.root = BTreeInteriorNode.from_JSON(btree, btreejson["root"])
		btree.root.isroot = True
		return btree

	def __str__(self):
		return str(self.root)

	def __getitem__(self, key):
		leaf = self.root.search(key)
		return leaf[key]

	def __setitem__(self, key, value):
		leaf = self.root.search(key)
		leaf[key] = value

	def __contains__(self, key):
		leaf = self.root.search(key)
		return key in leaf

	def _get_page_size(self):
		return self._page_size

	def _get_leaffactory(self):
		return self._leaffactory

	def _get_value_class(self):
		return self._value_class

	def _get_key_class(self):
		return self._key_class

	def insert(self, key, value):
		leaf = self.root.search(key)
		leaf.insert(key, value)

	def get_leafs(self):
		return self.root.get_leafs()

	@abstractmethod
	def remove(self, key):
		leaf = self.root.search(key)
		leaf.remove(key)

	page_size = property(_get_page_size)
	leaffactory = property(_get_leaffactory)
	key_class = property(_get_key_class)
	value_class = property(_get_value_class)


class BTreeInteriorNode(BTreeNode):
	def __init__(self, parent):
		super(BTreeInteriorNode, self).__init__(parent)
		self.keys = []
		self.childs = []
		self.isroot = False

	def __len__(self):
		return len(self.keys)

	def __str__(self):
			return "[" + str(self.childs[0]) + "".join([str(self.keys[i]) + str(self.childs[i+1]) for i in range(0,len(self))]) + "]"

	def to_JSON(self):
		nodejson = {}
		if self.key_class in [int, dict, list, str, str]:
			nodejson["keys"] = self.keys
		else:
			nodejson["keys"] = [key.to_JSON() for key in self.keys]

		nodejson["childs"] = [child.to_JSON() for child in self.childs]
		return nodejson

	@staticmethod
	def from_JSON(parent, nodejson):
		node = BTreeInteriorNode(parent)
		if parent.key_class in [int, dict, list, str, str]:
			node.keys = nodejson["keys"]
		else:
			node.keys = [parent.key_class.from_JSON(keyjson) for keyjson in nodejson["keys"]]

		node.childs = [BTreeNode.from_JSON(node, childjson) for childjson in nodejson["childs"]]
		return node

	def before(self, key_index):
		return self.childs[key_index]

	def after(self, key_index):
		return self.childs[key_index+1]

	def insert(self, key, node):
		#insert
		index = 0
		while index < len(self) and self.keys[index] <= key:
			index += 1
		self.keys.insert(index, key)
		self.childs.insert(index+1, node)
		node.parent = self

		#balance
		if len(self) > self.page_size:
			key, newnode = self.split()

			if self.isroot: #root was split, create new root
				newroot = BTreeInteriorNode(self.parent)
				newroot.isroot = True
				newroot.childs.append(self)
				newroot.insert(key, newnode)
				self.parent.root = newroot
				self.parent = newroot
				self.isroot = False
			else:
				self.parent.insert(key, newnode)

	def split(self):
		center = len(self) // 2
		key = self.keys[center]

		newnode = BTreeInteriorNode(None)
		newnode.keys = self.keys[center+1:]
		newnode.childs = self.childs[center+1:]

		for child in newnode.childs:
			child.parent = newnode

		self.keys = self.keys[:center]
		self.childs = self.childs[:center+1]

		return key, newnode

	def merge(self, key, node):
		keys = []
		keys.append(key)
		keys.extend(node.keys)

		for (key, node) in zip(keys, node.childs):
			self.insert(key, node)

	def borrowLeft(self, key, node):
		keys = []
		keys.extend(node.keys)
		keys.append(key)
		pairs = list(zip(keys, node.childs))
		pairs.reverse()
		oldseperator = key
		while not(len(self) >= self.page_size // 2) or len(self) < len(node)-1:
			key, child = pairs.pop(0)
			self.keys.insert(0,key)
			self.childs.insert(0,child)
			child.parent = self
			del node.childs[-1]
			newseperator = node.keys[-1]
			del node.keys[-1]
		node.parent.replaceKey(oldseperator,newseperator)

	def borrowRight(self, key, node):
		keys = []
		keys.append(key)
		keys.extend(node.keys)
		pairs = list(zip(keys, node.childs))
		oldseperator = key
		newseperator = key
		while not(len(self) >= self.page_size // 2) or len(self) < len(node)-1:
			key, child = pairs.pop(0)
			self.keys.append(key)
			self.childs.append(child)
			child.parent = self
			del node.childs[0]
			newseperator = node.keys[0]
			del node.keys[0]
		node.parent.replaceKey(oldseperator,newseperator)

	def replaceKey(self, oldkey, newkey):
		node = self
		while not(isinstance(node, BTree)):
			if oldkey in node.keys:
				keyindex = node.keys.index(oldkey)
				del node.keys[keyindex]
				node.keys.insert(keyindex, newkey)
			node = node.parent

	def remove(self, node):
		#delete
		index = self.childs.index(node)

		if index > 0:
			if index == 1 and len(self) > 1:
				oldkey = self.keys[0]
				newkey = self.keys[1]
				parent = self.parent
				while not(isinstance(parent, BTree)):
					if oldkey in parent.keys:
						keyindex = parent.keys.index(oldkey)
						del parent.keys[keyindex]
						parent.keys.insert(keyindex, newkey)
					parent = parent.parent

			del self.keys[index-1]

		del self.childs[index]

		if (isinstance(node, BTreeLeafNode)):
			self.leaffactory.deallocate_leaf(node)

		#balance
		if len(self) < self.page_size//2:
			if self.isroot:
				#collapse root
				no_keys = len(self) + sum([len(child) for child in self.childs])
				if no_keys <= self.page_size and (isinstance(self.childs[0], BTreeInteriorNode)):
						self.childs[0].parent = self.parent
						for (key,node) in zip(self.keys, self.childs[1:]):
							self.childs[0].merge(key,node)
						self.parent.root = self.childs[0]
						self.parent.root.isroot = True
			else:
				#borrow key from neighbor, try right first
				index = self.parent.childs.index(self)
				if index+1 < len(self.parent.childs) and len(self.parent.childs[index+1]) > self.page_size // 2:
					node = self.parent.childs[index+1]
					key = self.parent.keys[index]
					self.borrowRight(key,node)
					return

				if index-1 >= 0 and len(self.parent.childs[index-1]) > self.page_size // 2:
					node = self.parent.childs[index-1]
					key = self.parent.keys[index-1]
					self.borrowLeft(key,node)
					return

				#merge with neighbor, try right first
				if index+1 < len(self.parent.childs):
					node = self.parent.childs[index+1] #try right first
					key = self.parent.keys[index]
					self.merge(key, node)
					self.parent.remove(node)
					return

				if index-1 >= 0:
					node = self.parent.childs[index-1] #merge with left instead
					key = self.parent.keys[index-1]
					node.merge(key, self)
					self.parent.remove(self)
					return

	def get_leafs(self, key = None):
		leafs = self.childs[0].get_leafs(None)
		for i in range(1,len(self.childs)):
			leafs.extend(self.childs[i].get_leafs(self.keys[i-1]))

		return leafs




# coding=utf-8
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import object
import queue, threading, time, sys, datetime, multiprocessing, signal, threading, sys, itertools

from . import operations

TERMINAL_WIDTH = 80

active_processes = []

def _signal_handler(signal, frame):
	"""Used internally. Handles interruptions (like pressing CTRL-C) and stops running processes.
	Args:
		signal: Signal
		frame: Stack frame"""
	global active_processes
	for proc in active_processes:
		print("Aborting " + proc.name + " (can take a minute)...")
		proc.stop()

class Dataset(operations.Source):
	"""Represents a dataset (can be used as a source, intermediate dataset or result)"""
	def __init__(self, source, length=None, name=None):
		"""Creates a new dataset
		Args:
			source: A source where the data for this dataset is obtained from. An iterable or `pyparade.operations.Source`
			length: The number of elements in the dataset. Only used, if len(source) is not available
			name: The display name for the dataset (convention: 1-25 characters starting with a big letter)"""
		super(Dataset, self).__init__(name)

		self.source = source
		self._length = length

		try:
			self._length = len(source)
		except Exception as e:
			pass
		if self._length != None:
			self._length_is_estimated = False
		else:
			self._length_is_estimated = True

		if self.name == None:
			if isinstance(self.source, operations.Source) and self.source.output_name != None: #get name from operation output
				self.name = self.source.output_name
			else:
				self.name = "Dataset" 

		self._buffers = []
		self.finished = threading.Event()

	def __len__(self):
		if self._length != None:
			return self._length
		else:
			raise RuntimeError("Length is not available")

	def _get_buffer(self, size = 30):
		buf = Buffer(self, size=size)
		self._buffers.append(buf)
		return buf

	def _fill_buffers(self):
		self.running.set()
		if isinstance(self.source, operations.Operation):
			values = self.source()
		else:
			values = itertools.chain((value for value in self.source), [operations.OutputEndMarker()])

		if self._length_is_estimated:
			self._length = 0

		batch = []
		last_insert = 0

		for value in values:
			if self._check_stop():
				break

			while len([buf for buf in self._buffers if buf.full()]) > 0:
				if self._check_stop():
					break
				time.sleep(1)

			if self._check_stop():
				break

			batch.append(value)

			if self._length_is_estimated:
				if not isinstance(value, operations.OutputEndMarker):
					self._length += 1

			if time.time() - last_insert > 0.5:
				[buf.put(batch) for buf in self._buffers]
				batch = []
				last_insert = time.time()

		while len([buf for buf in self._buffers if buf.full()]) > 0:
			if self._check_stop():
				break
			time.sleep(1)

		[buf.put(batch) for buf in self._buffers]

		self._length_is_estimated = False
		self.finished.set()
		self.running.clear()

	def has_length(self):
		"""Returns True if the length of this Source is available. 
		If this function returns False, calling len() on the Source will raise a RuntimeException."""
		return self._length != None

	def length_is_estimated(self):
		"""Returns True if the length of this Source is estimated and might change over time,
		for example if not all items in the Source are known yet."""
		return self._length_is_estimated

	def map(self, map_func, context = None, **kwargs):
		"""Returns a new `pyparade.Dataset` which results from applying map_func to each element in this Dataset.

		Supplying a context can be used to establish a connection to a common resource such as a database.

		Args:
			map_func: The function to be applied to each element. Has to accept an element from this Dataset as the first 
					  argument and should return a new element which is put into the output Dataset.
			context: A function that returns a context manager. It is called once for each parallel executor 
					 which executes map_func. For each call of map_func the context object is passed as the second argument.
			**kwargs: Other arguments are passed on to `pyparade.operations.MapOperation

		Example:
			>>> import pyparade
			>>> d = pyparade.Dataset([1,2,3])
			>>> d.map(lambda a: a + 1).collect() #add 1 to each element
			[2,3,4]

		"""
		op = operations.MapOperation(self, map_func, context = context, **kwargs)
		return Dataset(op)

	def flat_map(self, map_func, context = None, **kwargs):
		"""Returns a new `pyparade.Dataset` which results from applying map_func to each element in 
		this Dataset and combining the returned lists in a flat list.

		Supplying a context can be used to establish a connection to a common resource such as a database.

		Args:
			map_func: The function to be applied to each element. Has to accept an element from this Dataset as the first 
					  argument and should return an iterable containing elements which are put into the output Dataset.
			context: A function that returns a context manager. It is called once for each parallel executor 
					 which executes map_func. For each call of map_func the context object is passed as the second argument.
			**kwargs: Other arguments are passed on to `pyparade.operations.FlatMapOperation`

		Example:
			>>> import pyparade
			>>> d = pyparade.Dataset(["This is a test", "a b c"])
			>>> d.map(str.split).collect() #split by space (map)
			[["This", "is", "a", "test"], ["a", "b", "c"]]
			>>> d.flat_map(str.split).collect() #split by space and flat map
			["This", "is", "a", "test", "a", "b", "c"]
		"""
		op = operations.FlatMapOperation(self, map_func, context = context, **kwargs)
		return Dataset(op)

	def batch(self, batch_size=1, **kwargs):
		"""Returns a new `pyparade.Dataset` containing the elements of this dataset in batches (lists of equal length)

		Args:
			batch_size: The number of elements to put in each batch
			**kwargs: Other arguments are passed on to `pyparade.operations.BatchOperation

		Example:
			>>> import pyparade
			>>> d = pyparade.Dataset(1,2,3,4,5)
			>>> d.batch(2).collect()
			[[1,2],[3,4],[5]]

		"""
		op = operations.BatchOperation(self, batch_size=batch_size, **kwargs)
		return Dataset(op)

	def group_by_key(self, partly = False, **kwargs):
		"""Returns a new `pyparade.Dataset` which results from grouping (key,value) tuples by their key into tuples (key, [values]).

		Args:
			partly: If True, partial groups can be returned. This allows streaming processing with output 
					starting before all elements in the dataset have been processed.
			**kwargs: Other arguments are passed on to `pyparade.operations.GroupByKeyOperation

		Example:
			>>> import pyparade
			>>> d = pyparade.Dataset([("a", 1), ("b", 1), ("a",2)])
			>>> d.group_by_key().collect()
			[("a", [1,2]), ("b", [1])]

		"""
		op = operations.GroupByKeyOperation(self, partly = partly, **kwargs)
		return Dataset(op)

	def reduce_by_key(self, reduce_func, **kwargs):
		"""Returns a new `pyparade.Dataset` which results from grouping (key,value) tuples by their key into tuples (key, [values]) 
		and applying reduce_func to each (key,[values]) tuple.

		Args:
			reduce_func: The function to be applied to each (key,[values]) tuple. Has to accept a (key,[values]) tuple as the first 
					  argument and should return a new element which is put into the output Dataset.
			**kwargs: Other arguments are passed on to `pyparade.operations.ReduceByKeyOperation`
		"""
		op = operations.ReduceByKeyOperation(self, reduce_func, **kwargs)
		return Dataset(op)

	def fold(self, zero_value, fold_func, context = None, **kwargs):
		"""Returns a new `pyparade.Dataset` containing one element which results by repeatedly applying the fold function.

		Supplying a context can be used to establish a connection to a common resource such as a database.

		Args:
			zero_value: An intial value that represents zero.
			fold_func: The function to be used to fold two values. Must accept two values as arguments.
					   Must return a new value that is again accepted as an argument to fold_func.
			context: A function that returns a context manager. It is called once for each parallel executor 
					 which executes map_func. For each call of map_func the context object is passed as the second argument.
			**kwargs: Other arguments are passed on to `pyparade.operations.FoldOperation`

		Example:
			>>> import pyparade, operator
			>>> d = pyparade.Dataset([1,2,3])
			>>> d.fold(0, operator.add).collect() #sum elements
			[6]
		"""
		op = operations.FoldOperation(self, zero_value, fold_func, context = context, **kwargs)
		return Dataset(op)

	def start_process(self, name="Parallel Process", num_workers=multiprocessing.cpu_count(), **kwargs):
		"""Starts and returns a `pyparade.ParallelProcess` to collect elements in this dataset. 
		Normally called indirectly using `pyparade.Dataset.collect`.

		Args:
			name: Name of the parallel process
			num_workers: Number of parallel workers to use (default: number of available CPUs)  
			kwargs: remaining keyword arguments are passed on to ParallelProcess"""
		proc = ParallelProcess(self, name, **kwargs)
		proc.run(num_workers)
		return proc

	def _stop_process(self, process, old_handler):
		"""Internal method to stop a parellel process.

		Args:
			name: Name of the parallel process
			num_workers: Number of parallel workers to use (default: number of available CPUs)  """
		global active_processes

		if old_handler != None:
			signal.signal(signal.SIGINT, old_handler)
		else:
			signal.signal(signal.SIGINT, signal.SIG_DFL)

		if process in active_processes:
			active_processes.remove(process)

		if threading.active_count() > 1:
			time.sleep(2)

		while threading.active_count() > 1:
			print("Hanging threads:")
			for t in threading.enumerate():
				if t.isAlive() and not(t == threading.current_thread()):
					print(t.name)
			time.sleep(5)
		raise RuntimeError("Process was stopped")

	def collect(self, **args):
		"""Returns a list of all elements in this dataset. 
		Starts a `ParallelProcess` in order to collect the data.

		Args:
			**args: All arguments are passed on to `pyparade.Dataset.start_process` """
		global active_processes

		old_handler = None
		proc = None
		if not self.running.is_set(): #no process running yet, start process
			proc = self.start_process(**args)
			active_processes.append(proc)
			old_handler = signal.getsignal(signal.SIGINT)
			signal.signal(signal.SIGINT, _signal_handler) #abort on CTRL-C

		if self._stop_requested.is_set():
			self._stop_process(proc, old_handler)

		result = []
		for val in self._get_buffer().generate():
			if self._stop_requested.is_set():
				self._stop_process(proc, old_handler)

			result.append(val)

		if self._stop_requested.is_set():
			self._stop_process(proc, old_handler)

		if old_handler != None:
			signal.signal(signal.SIGINT, old_handler)
		else:
			signal.signal(signal.SIGINT, signal.SIG_DFL)
		
		if proc in active_processes:
			active_processes.remove(proc)

		return result

class ParallelProcess(object):
	"""A parallel process that collects data in a `pyparade.Dataset`"""
	def __init__(self, dataset, name="Parallel process", status=True, status_interval=15):
		"""Creates a new parallel process
		Args:
			dataset: The `pyparade.Dataset` which the process should collect
			name: The display name for the process
			print_status: Display status information during processing?
			print_status_interval: Update interval of status information in seconds
		"""

		self.dataset = dataset
		self.result = []
		self.name = name
		self.status = status
		self.status_interval = status_interval

	def run(self, num_workers = multiprocessing.cpu_count()):
		#Build process tree
		chain = self.dataset.get_parents()
		chain.reverse()
		self.chain = chain

		for source in self.chain:
			source.processes.append(self)

		#set number of workers
		for operation in [block for block in chain if isinstance(block, operations.Operation)]:
			operation.num_workers = num_workers

		threads = []
		for dataset in [block for block in chain if isinstance(block, Dataset)]:
			t = threading.Thread(target = dataset._fill_buffers, name="Buffer")
			t.start()
			threads.append(t)

		if self.status:
			ts = threading.Thread(target = self.print_status)
			ts.start()

	def stop(self):
		[s.stop() for s in self.chain]

	def clear_screen(self):
		"""Clear screen, return cursor to top left"""
		sys.stdout.write('\033[2J')
		sys.stdout.write('\033[H')
		sys.stdout.flush()

	def print_status(self):
		started = time.time()

		#self.clear_screen()
		#print(self.get_status())
		t = 0
		while not len([s for s in self.chain if s.finished.is_set()]) == len(self.chain):
			try:
				time.sleep(1)
				t += 1
				if t >= self.status_interval:
					self.clear_screen()
					print(self.get_status())
					t = 0
			except Exception as e:
				print(e)
				time.sleep(60)
		self.clear_screen()
		print(self.get_status())

		ended = time.time()
		print("Computation took " + str(ended-started) + "s.")

	def get_status(self):
		txt = util.shorten(str(self), TERMINAL_WIDTH) + "\n"
		txt += ("=" * TERMINAL_WIDTH) + "\n"
		txt += "\n".join([self.get_buffer_status(op) + "\n" + self.get_operation_status(op) for op in self.chain if isinstance(op, operations.Operation)])
		txt += "\n" + self.get_result_status()
		return txt

	def get_buffer_status(self, op):
		status = ""

		if not op.source.length_is_estimated():
			status += str(len(op.source))
		elif not op.source.running.is_set():
			status += "stopped"

		title = "" 
		if len(op.inbuffer) > 0:
			title += " " + "(buffer: " + str(len(op.inbuffer)) + ")"
		title = util.shorten(str(op.source), TERMINAL_WIDTH - len(title) - len(status)) + title
		space = " "*(TERMINAL_WIDTH - len(title) - len(status))
		return title + space + status

	def get_operation_status(self, op):
		status = ""

		if op.exception != None:
			status += "FAILED"
		elif op._check_stop():
			status += "stopping"
		else:
			if op.source.has_length():
				if not op.source.length_is_estimated() and len(op.source) > 0 and op.processed > 0:
					if op.processed == len(op.source):
						status += "done"
					elif op.running.is_set():
						est = datetime.datetime.now() + datetime.timedelta(seconds = (time.time()-op.time_started)/op.processed*(len(op.source)-op.processed))
						status += '{0:%}'.format(float(op.processed)/len(op.source)) + "  ETA " + est.strftime("%Y-%m-%d %H:%M") + " "
						status += str(op.processed) + "/" + str(len(op.source))
					else:
						status += "stopped"
				else:
					status += str(op.processed) + "/" + str(len(op.source))
			else:
				if not op.running.is_set():
					status += "stopped"
			


		title = util.shorten(str(op), (TERMINAL_WIDTH - len(str(op)) - len(status) - 2))
		space = " "*(TERMINAL_WIDTH - len(str(title)) - len(status) - 1)
		return " " + title + space + status

	def get_result_status(self):
		status = ""
		if self.dataset.has_length():
			status = str(len(self.dataset))

		title = " (result)"
		title = util.shorten(str(self.dataset), TERMINAL_WIDTH - len(title) - len(status)) + title
		space = " "*(TERMINAL_WIDTH - len(title) - len(status))
		return title + space + status

	def __str__(self):
		return self.name

class Buffer(object):
	"""A thread-safe buffer used to read buffered from a `pyparade.operations.Source`"""
	def __init__(self, source, size):
		"""Creates a new Buffer.

		Args:
			source: The `pyparade.operations.Source` which the buffer reads from
			size: The size of the buffer (number of elements to keep in the buffer) """

		super(Buffer, self).__init__()
		self.source = source
		self.size = size
		self.queue = queue.Queue(size)
		self._length = 0
		self._length_lock = threading.Lock()

	def __len__(self):
		with self._length_lock:
			return self._length

	def full(self):
		"""Returns True if the buffer is full, that is the number of elements in the buffer is equal to the buffer size."""
		return self.queue.full()

	def put(self, values):
		"""Puts a batch of elements into the buffer

		Args:
			values: An iterable containing the elements to but into the buffer. Must support calling len(values)."""

		self.queue.put(values, True)

		if not (len(values) > 0 and isinstance(values[-1], operations.OutputEndMarker)):
			with self._length_lock:
				self._length += len(values)
		else:
			print("test")
			with self._length_lock:
				self._length += len(values)-1

	def generate(self):
		"""A generator yielding elements from the buffer. Runs until the underlying `pyparade.operations.Source` is finished."""

		finished = False

		while not finished or self.source._stop_requested.is_set():
			try:
				values = self.queue.get(True, timeout=1)
				for value in values:
					if not isinstance(value, operations.OutputEndMarker):
						with self._length_lock:
							self._length -= 1
						yield value
					else:
						finished = True
			except Exception as e:
				pass

		chain = [self.source] + self.source.get_parents()
		chain.reverse()
		
		for s in chain:
			if s.exception:
				for p in self.source.processes:
					p.stop()
				ex_type, ex_value, tb_str = s.exception
				message = '%s (in %s)' % (str(ex_value), s.name)
				raise ex_type(message)



from past.builtins import cmp
from builtins import str
from builtins import range
from builtins import object
from functools import total_ordering

import os, random, tempfile, shutil
import unittest
import pkg_resources


from pyparade.util.btree import BTree, BTreeFileLeafFactory

class TestBTree(unittest.TestCase):
	def setUp(self):
		self.btree = BTree(3, int, int)

	def tearDown(self):
		pass

	def test_insert(self):
		self.btree.insert(1,101)
		self.assertEqual(1, self.btree.root.childs[0].keys[0])
		self.assertEqual(101, self.btree.root.childs[0].values[0])

		self.btree.insert(2,102)
		self.assertEqual(1, self.btree.root.childs[0].keys[0])
		self.assertEqual(101, self.btree.root.childs[0].values[0])
		self.assertEqual(2, self.btree.root.childs[0].keys[1])
		self.assertEqual(102, self.btree.root.childs[0].values[1])

		self.btree.insert(3,103)
		self.assertEqual([1,2,3], self.btree.root.childs[0].keys)
		self.assertEqual([101,102,103], self.btree.root.childs[0].values)

		self.btree.insert(4,104)
		self.assertEqual([1,2], self.btree.root.childs[0].keys)
		self.assertEqual([101,102], self.btree.root.childs[0].values)
		self.assertEqual([3,4], self.btree.root.childs[1].keys)
		self.assertEqual([103,104], self.btree.root.childs[1].values)
		self.assertEqual([3], self.btree.root.keys)

		numbers = [x for x in range(5,100)]
		random.shuffle(numbers)

		for number in numbers:
			self.btree.insert(number, 100+number)

		for number in numbers:
			self.assertIn(number, self.btree.root.search(number).keys)
			self.assertIn(100+number, self.btree.root.search(number).values)

	def test_remove(self):
		#insert
		numbers = [x for x in range(1,1000)]
		#numbers = [4, 13, 2, 12, 9, 10, 8, 1, 6, 5, 3, 7, 11, 14] 
		random.shuffle(numbers)

		for number in numbers:
			self.btree.insert(number, 100+number)

		#test delete
		random.shuffle(numbers)

		#numbers = [10,6,14,12,9,4,5,3,1,13,8,11,2,7]
		for number in numbers:
			self.assertIn(number, self.btree.root.search(number).keys)
			self.assertIn(100+number, self.btree.root.search(number).values)
			self.btree.remove(number)
			self.assertNotIn(number, self.btree.root.search(number).keys)
			self.assertNotIn(100+number, self.btree.root.search(number).values)

		self.assertEqual(0, len(self.btree.root))
		self.assertEqual([], self.btree.root.childs[0].keys)
		self.assertEqual([], self.btree.root.childs[0].values)

@total_ordering
class IntClass(object):
	def __init__(self, integer):
		self.integer = integer

	def to_JSON(self):
		return self.integer

	@staticmethod
	def from_JSON(json):
		return IntClass(json)

	def __str__(self):
		return str(self.integer)

	def __lt__(self, other):
		return self.integer < other.integer

	def __eq__(self, other):
		return (isinstance(other, self.__class__)) and (self.integer == other.integer)

class TestBTreeFileLeaves(unittest.TestCase):
	def setUp(self):
		self.tempdir = tempfile.mkdtemp()
		self.leaffactory = BTreeFileLeafFactory(os.path.join(self.tempdir, "page"), ".index")
		self.btree = BTree(50, IntClass, IntClass, self.leaffactory)
	
	def tearDown(self):
		shutil.rmtree(self.tempdir)
		pass

	def test_insert(self):
		numbers = [x for x in range(1,1000)]
		random.shuffle(numbers)

		for number in numbers:
			self.btree.insert(IntClass(number), IntClass(100+number))
			self.assertIn(number, [key.integer for key in self.btree.root.search(IntClass(number)).keys])
			self.assertIn(100+number, [value.integer for value in self.btree.root.search(IntClass(number)).values])
			self.leaffactory.cleanup()
		self.assertGreater(len(self.leaffactory.allocated_leaves),2)

	def test_remove(self):
		numbers = [x for x in range(1,1000)]
		random.shuffle(numbers)
		#insert
		for number in numbers:
			self.btree.insert(IntClass(number), IntClass(100+number))
			self.assertIn(number, [key.integer for key in self.btree.root.search(IntClass(number)).keys])
			self.assertIn(100+number, [value.integer for value in self.btree.root.search(IntClass(number)).values])

		#test delete
		random.shuffle(numbers)
		for number in numbers:
			leaf = self.btree.root.search(IntClass(number))
			leaf.ensure_load()
			self.assertIn(number, [key.integer for key in self.btree.root.search(IntClass(number)).keys])
			self.assertIn(100+number, [value.integer for value in self.btree.root.search(IntClass(number)).values])

			self.btree.remove(IntClass(number))

			leaf = self.btree.root.search(IntClass(number))
			leaf.ensure_load()
			self.assertNotIn(number, [key.integer for key in self.btree.root.search(IntClass(number)).keys])
			self.assertNotIn(100+number, [value.integer for value in self.btree.root.search(IntClass(number)).values])
			self.leaffactory.cleanup()

		self.assertEqual(0, len(self.btree.root))
		self.assertEqual([], self.btree.root.childs[0].keys)
		self.assertEqual([], self.btree.root.childs[0].values)
		self.assertEqual(1, len(self.leaffactory.allocated_leaves))

	def test_persistent(self):
		#load with numbers
		numbers = [x for x in range(1,1000)]
		random.shuffle(numbers)

		for number in numbers:
			self.btree.insert(IntClass(number), IntClass(100+number))

		#serialize and deserialize again
		treejson = self.btree.to_JSON()
		self.leaffactory = BTreeFileLeafFactory(os.path.join(self.tempdir, "page"), ".index")
		self.btree = BTree.from_JSON(treejson, IntClass, IntClass, self.leaffactory)

		#check for numbers
		for number in numbers:
			leaf = self.btree.root.search(IntClass(number))
			leaf.ensure_load()
			self.assertIn(number, [key.integer for key in leaf.keys])

		self.assertGreater(len(self.leaffactory.allocated_leaves),2)

		#test delete to check if parent correctly set on all nodes and tree works
		random.shuffle(numbers)
		for number in numbers:
			leaf = self.btree.root.search(IntClass(number))
			leaf.ensure_load()
			self.assertIn(number, [key.integer for key in self.btree.root.search(IntClass(number)).keys])
			self.assertIn(100+number, [value.integer for value in self.btree.root.search(IntClass(number)).values])

			self.btree.remove(IntClass(number))

			leaf = self.btree.root.search(IntClass(number))
			leaf.ensure_load()
			self.assertNotIn(number, [key.integer for key in self.btree.root.search(IntClass(number)).keys])
			self.assertNotIn(100+number, [value.integer for value in self.btree.root.search(IntClass(number)).values])
			self.leaffactory.cleanup()

		self.assertEqual(0, len(self.btree.root))
		self.assertEqual([], self.btree.root.childs[0].keys)
		self.assertEqual([], self.btree.root.childs[0].values)
		self.assertEqual(1, len(self.leaffactory.allocated_leaves))




from setuptools import setup, find_packages
import sys, os

version = '0.4.3'

setup(name='pyparade',
      version=version,
      description="Python library to parallel process data efficiently",
      long_description="""\
PyParade is a lightweight multiprocessing framework that makes it easy to parallel process large-scale datasets efficiently.""",
      classifiers=["License :: OSI Approved :: MIT License",
                   "Development Status :: 3 - Alpha",
                   "Environment :: Console",
                   "Intended Audience :: Information Technology",
                   "Natural Language :: English",
                   "Operating System :: MacOS :: MacOS X",
                   "Operating System :: POSIX :: Linux",
                   "Programming Language :: Python"],
      keywords='big data parallel processing',
      author='Nils Breyer',
      author_email='mail@nilsbreyer.eu',
      url='http://nilsbreyer.eu/#!projects/pyparade',
      license='MIT License',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=True,
      install_requires=[
            'future>=0.18.2'
      ]
      )